/*
 * XCAPServer.java
 *
 * Created on August 1, 2004, 10:07 AM
 * Author - Madhuri Shinde
 *email - mss2103@columbia.edu
 */

package edu.columbia.xcap;

import java.io.*;
import java.net.*;
import java.lang.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import java.security.*;
// import the required classes needed for the DOM parsing
import org.w3c.dom.*;
import org.apache.xerces.dom.*;
import org.apache.xerces.parsers.*;
import org.apache.xml.serialize.*;
import org.xml.sax.*;
import org.apache.commons.httpclient.auth.*; 

/**
 *
 * @author  Madhuri
 * @version
 */
public class XCAPServer extends HttpServlet {
    // the root directory where all the applications are stored
    public String ROOT_DIR;
    // the index file which is global is is inorporated in all applications
    public String INDEX_FILE;
    
    /** Initializes the servlet.
     */
    public void init(ServletConfig config) throws ServletException {
        super.init(config);
        // set the root to a folder on local machine
        // for clic machine
        ROOT_DIR = "/home/mss2103/xcap/resources";
        // for windows machine
        //ROOT_DIR = "/XCAP";
        // global index file
        INDEX_FILE = "index.xml";
    }
    
    /** Destroys the servlet.
     */
    public void destroy() {
        
    }
        
    protected void configParser(DOMParser parser) {
      try {
          parser.setFeature("http://apache.org/xml/features/dom/defer-node-expansion", false); // otherwise parser.getCurrentNode() == null
          parser.setFeature("http://apache.org/xml/features/continue-after-fatal-error", false);
          parser.setFeature("http://apache.org/xml/features/allow-java-encodings", true);
          
          parser.setFeature("http://xml.org/sax/features/validation", true);
          parser.setFeature("http://xml.org/sax/features/namespaces", true);
          parser.setFeature("http://apache.org/xml/features/validation/schema", true);
          parser.setFeature("http://apache.org/xml/features/validation/schema-full-checking", true);
          parser.setFeature("http://apache.org/xml/features/validation/dynamic", true);          
      }
      catch (Exception e) {
      }       
    }    
    
    /**
     * Check for well-formatting and schema-validation. On error it throws the exception.
     * Returns the parsed document object on success.
     */
    protected Document getParsedDocument(String schemafile, InputSource input)
    throws IOException, SAXException {
        // Create and configure a DOM parser
        DOMParser parser = new DOMParser();

        // configure the parser
        configParser(parser);

        File schema = new File(schemafile);
        String targetNameSpace = "urn:ietf:params:xml:ns:conference-policy";
        if (schema.exists()) {
          //TBD: open the schema file, get the target namespace and use that.
          parser.setProperty("http://apache.org/xml/properties/schema/external-schemaLocation",
              targetNameSpace
              + " file://" + schemafile);
          //This ErrorHandler also needs to be installed with the parser. 
          parser.setErrorHandler(new ErrorsAreFatal());
        }
        else {
            System.out.println("Schema file does not exist: " + schemafile);
        }

        // Parse the document
        parser.parse(input);
        
        Document theDocument = parser.getDocument(); // the document

        if (schema.exists()) {
            Element topElement = theDocument.getDocumentElement();
            if (topElement == null) {
                throw new SAXException("No top level element.");
            }

            String ns = topElement.getAttribute("xmlns");
            if (ns == null || !ns.equals(targetNameSpace)) {
                throw new SAXException("Not in valid name space");
            }
        }
        
        System.out.println("Parsed correctly\n");
        return theDocument;
     }
    
     /**
      * The schema file is present as application/global/schema.xsd
      */
     protected String getSchemaFileName(String appName) 
     {
         return new String(ROOT_DIR + "/" + appName + "/" + "global/schema.xsd");
     }

     /**
      * The content type is used as "application/appname+xml" for any given
      * application name "appname".
      */
     protected String getApplicationContentType(String appName)
     {
       String type = "application/" + appName + "+xml";
       return type;
     }
     
     /**
      * Get the whole text in a string from the bufferedReader stream. It closes the
      * reader stream before returning.
      */
     protected String getTextFromReader(BufferedReader in)
     throws IOException {
        String body = new String("");
        String str;
        while ((str = in.readLine()) != null) {
            body += str;
            body += "\r\n";
        }
        in.close();
        return body;
     }

     /**
      * Put a text string to the print writer stream. Closes the stream before
      * returning.
      */
     protected void putTextToWriter(PrintWriter out, String body)
     {
        out.write(body);
        out.close();
     }
  
     /**
      * GET for the whole document.
      */
     protected void doGetDocument (HttpServletResponse response, String appName, File document) 
     throws IOException {
        try {
            response.setContentType(getApplicationContentType(appName));

            String body = getTextFromReader(new BufferedReader(new FileReader(document)));
            String schemafile = getSchemaFileName(appName);

            Document theDocument = getParsedDocument(schemafile, new InputSource(new StringReader(body)));
            putTextToWriter(response.getWriter(), body);
        }
        catch (SAXException e) {
            response.sendError(HttpServletResponse.SC_CONFLICT, "Error parsing the XML document: " + e.getMessage());
            return;
        }
        catch (IOException e) {
            setLastModified(response, document);
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Error reading file: " + e.getMessage());
            return;
        }        
        catch (Exception e) {
            response.sendError(HttpServletResponse.SC_CONFLICT, "Error in the document: " + e.getMessage());
            return;
        } 
     }    

     /**
      * PUT for the whole document.
      */
     protected void doPutDocument (HttpServletRequest request, HttpServletResponse response, 
                                   String appName, File document) 
     throws IOException {
        try {
            //check if the content type is right or not.
            String type = getApplicationContentType(appName);
            String r_type = request.getContentType();
            if ((r_type == null) || !type.equalsIgnoreCase(r_type)) {
                response.setHeader("Accept", type);
                response.sendError(HttpServletResponse.SC_UNSUPPORTED_MEDIA_TYPE, "Content type should be " + type);  
                return;     
            }

            String body = getTextFromReader(request.getReader());
            String schemafile = getSchemaFileName(appName);
            
            Document theDocument = getParsedDocument(schemafile, new InputSource(new StringReader(body)));
            putTextToWriter(new PrintWriter(new FileWriter(document)), body);
            
        } 
        catch (SAXException e) {
            response.sendError(HttpServletResponse.SC_CONFLICT, "Error parsing the XML document: " + e.getMessage());
        }
        catch (IOException e) {
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Error writing file: " + e.getMessage());
        }
        catch (Exception e) {
            response.sendError(HttpServletResponse.SC_CONFLICT, "Error in the document: " + e.getMessage());
        } 
    }    

    /**
     * Get the specified node from the document. On error it returns null and
     * sends the error response.
     */
    protected Node getNodeFromSelector(Document theDocument, String nodeSelector, HttpServletResponse response)
    throws IOException {
        Node node = theDocument; //set the top level to document
        
        String errorStep = "";
        String steps[] = nodeSelector.split("/");
        for (int j=0; j<steps.length; ++j) { // get the children of the present top level
            String step = steps[j];
            
            if (step != null && !step.startsWith("@")) {
                /* This must be an element selector. */
                // now split the step into element, position, attribute name, value, etc.
                if (!step.matches("[^\\[\\s]+((\\[\\d+\\])(\\[@[^\\s=]+=\"[^\"]*\"\\])??|(\\[@[^\\s=]+=\"[^\"]*\"\\]))??")) {
                    response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Syntax error in node selector step " + step);
                    return null;
                }

                String params[] = step.split("[\\[\\]]", 6);
                String elementName = params[0];
                
                Vector nodes = new Vector();
                NodeList children = node.getChildNodes();

                // for each of the child found, compare it with the path element
                for (int k=0, posCount=0; k<children.getLength(); ++k) { 
                   Node child = children.item(k);
                   if (child.getNodeType() == Node.ELEMENT_NODE) {
                       Element el = (Element) child;
                       if (elementName.equals("*") || elementName.equals(el.getTagName())) { 
                           // if match found, check that child for the right position, attribute etc

                            int elementPos = -1;
                            
                            int posIndex  = 1, attrIndex = 1;
                            if (params.length > posIndex && !params[posIndex].startsWith("@")) {
                                // then its position
                                try {
                                    elementPos = Integer.parseInt(params[1]);
                                }
                                catch (NumberFormatException nfe) {// occurs if the part cannnot be parsed, if it is not Integer
                                    response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Position " + params[1] + " is not an integer");                           
                                    return null;
                                }
                                attrIndex += 2;
                                
                                if (posCount != elementPos) {
                                    posCount++;
                                    continue; //go to the next element.
                                }
                                else {
                                    posCount++;
                                }
                            }
                            
                            if (params.length > attrIndex && params[attrIndex].startsWith("@")) {
                                String hv = params[attrIndex];
                                String attrParts[] = hv.substring(1, hv.length()).split("=");
                                String attrName = attrParts[0];
                                String attrValue = attrParts[1];
                                
                                if (!el.hasAttribute(attrName)) {
                                    if (elementPos >= 0) {
                                        response.sendError(HttpServletResponse.SC_NOT_FOUND, "Attribute " + attrName + " not found in the element.");
                                        return null;
                                    }
                                    else {
                                        continue;
                                    }
                                }
                                String attr = el.getAttribute(attrName);
                                if (!attrValue.equals("\"" + attr + "\"")) {
                                    if (elementPos >= 0) {
                                        response.sendError(HttpServletResponse.SC_NOT_FOUND, "Attribute " + attrName + " does not match in the element");
                                        return null;
                                    }
                                    else {
                                        continue;
                                    }
                                }
                            }
                           
                            node = child;
                            nodes.add(step);
                       }
                       else {
                           errorStep = step;
                       }
                   }
                }
                
                if (nodes.size() == 0) { // break the loop if some intermediate element was not found
                    response.sendError(HttpServletResponse.SC_NOT_FOUND, "Node selector not found for " + errorStep);
                    return null;
                }
                if (nodes.size() > 1) {
                    response.sendError(HttpServletResponse.SC_MULTIPLE_CHOICES, "Many matches found for " + errorStep);
                    return null;
                }   
            }
            else {
                /* this is an attribute selector. */
                //should also check that this is the last step
                if (j != steps.length - 1) {
                    response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Attribute selector must be the last step");
                    return null;
                }
                String attrName = step.substring(1, step.length());
                Element el = (Element) node;
                if (!el.hasAttribute(attrName)) {
                    response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Attribute not found for the element "+errorStep);
                    return null;
                }    
                node = el.getAttributeNode(attrName);
                break;
            }
        }
        
        return node;
    }
    
    protected Node[] getNodeFromSelector2(Document theDocument, String nodeSelector, HttpServletResponse response)
    throws IOException {
        Node node = theDocument; //set the top level to document
        //Element theDocTop = theDocument.getDocumentElement();
        Node parent = null;
        
        String errorStep = "";
        String steps[] = nodeSelector.split("/");
        for (int j=0; j<steps.length; ++j) { // get the children of the present top level
            String step = steps[j];
            errorStep.concat(steps[j]);
            if (step != null && !step.startsWith("@")) {
                // This must be an element selector.
                // now split the step into element, position, attribute name, value, etc.
                if (!step.matches("[^\\[\\s]+((\\[\\d+\\])(\\[@[^\\s=]+=\"[^\"]*\"\\])??|(\\[@[^\\s=]+=\"[^\"]*\"\\]))??")) {
                    response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Syntax error in node selector step " + step);
                    return null;
                }

                String params[] = step.split("[\\[\\]]", 6);
                String elementName = params[0];
                
                Vector nodes = new Vector();
                NodeList children = node.getChildNodes();
                parent = node;

                int elementPos = -1;                            
                if (params.length > 1 && !params[1].startsWith("@")) {
                    try {
                        elementPos = Integer.parseInt(params[1]);
                    }
                    catch (NumberFormatException nfe) {// occurs if the part cannnot be parsed, if it is not Integer
                        response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Position " + params[1] + " in " +errorStep+ " is not an integer");
                        return null;
                    }
                }
                
                int posCount = 0;
                
                // for each of the child found, compare it with the path element
                for (int k=0; k<children.getLength(); ++k) { 
                   Node child = children.item(k);
                   if (child.getNodeType() == Node.ELEMENT_NODE) {
                       Element el = (Element) child;
                       if (elementName.equals("*") || elementName.equals(el.getTagName())) { 
                           // if match found, check that child for the right position, attribute etc
                            int posIndex  = 1, attrIndex = 1;
                            if (params.length > posIndex && !params[posIndex].startsWith("@")) {
                                // then its position
                                attrIndex += 2;
                                
                                if (posCount != elementPos) {
                                    posCount++;
                                    continue; //go to the next element.
                                }
                                else {
                                    posCount++;
                                }
                            }
                            
                            if (params.length > attrIndex && params[attrIndex].startsWith("@")) {
                                String hv = params[attrIndex];
                                String attrParts[] = hv.substring(1, hv.length()).split("=");
                                String attrName = attrParts[0];
                                String attrValue = attrParts[1];
                                
                                if (!el.hasAttribute(attrName)) {
                                    if (elementPos >= 0) {
                                        if (j == steps.length - 1) {
                                            /* This is not an error for PUT, this is positional insert. */
                                            if (k <= children.getLength() - 1) {
                                               Node newNode = child.cloneNode(true);
                                               parent.insertBefore(newNode, child);
                                               nodes.add(step);
                                               node = newNode; //this will be replaced later.
                                            }
                                            break;
                                        }
                                        else {
                                            response.sendError(HttpServletResponse.SC_CONFLICT, "Attribute " + attrName + " not found in the element "+errorStep);
                                            return null;
                                        }
                                    }
                                    else {
                                        continue;
                                    }
                                }
                                String attr = el.getAttribute(attrName);
                                if (!attrValue.equals("\"" + attr + "\"")) {
                                    if (elementPos >= 0) {
                                        if (j == steps.length - 1) {
                                            /* This is not an error for PUT, this is positional insert. */
                                            if (k <= children.getLength() - 1) {
                                               Node newNode = child.cloneNode(true);
                                               parent.insertBefore(newNode, child);
                                               nodes.add(step);
                                               node = newNode; //this will be replaced later.
                                            }
                                            break;
                                        }
                                        else {
                                            response.sendError(HttpServletResponse.SC_CONFLICT, "Attribute " + attrName + " does not match in the element given by "+errorStep);
                                            return null;
                                        }
                                    }
                                    else {
                                        continue;
                                    }
                                }
                            }
                           
                            node = child;
                            nodes.add(step);
                       }
                   }
                }
                
                if (nodes.size() > 1) {
                    response.sendError(HttpServletResponse.SC_MULTIPLE_CHOICES, "Many matches found for " + errorStep);
                    return null;
                }
                if (elementPos >= 0 && elementPos > posCount) {
                      response.sendError(HttpServletResponse.SC_CONFLICT, "Not sufficient number of elements for " + errorStep);
                      return null;
                }
                if (nodes.size() == 0) { 
                    if (j == steps.length - 1) {
                        //add this element as child of original_node
                        node = null;
                        break;
                    }
                    else {
                      response.sendError(HttpServletResponse.SC_CONFLICT, "Node selector not found for " + errorStep);
                      return null;
                    }
                }
            }
            else {
                // this is an attribute selector.
                //should also check that this is the last step
                if (j != steps.length - 1) {
                    response.sendError(HttpServletResponse.SC_CONFLICT, "Attribute selector must be the last step");
                    return null;
                }
                String attrName = step.substring(1, step.length());
                Element el = (Element) node;
                parent = node;
                if (el.hasAttribute(attrName)) {
                    node = el.getAttributeNode(attrName);
                }    
                else {
                    node = null;
                }
                break;
            }
        }
        
        Node[] nodes = new Node[2];
        nodes[0] = parent;
        nodes[1] = node;
        return nodes;
    }
    
    protected Node[] getNodeFromSelector3(Document theDocument, String nodeSelector, HttpServletResponse response)
    throws IOException {
        Node node = theDocument; //set the top level to document
        //Element theDocTop = theDocument.getDocumentElement();
        Node parent = null;
        
        String errorStep = "";
        String steps[] = nodeSelector.split("/");
        for (int j=0; j<steps.length; ++j) { // get the children of the present top level
            String step = steps[j];
            
            if (step != null && !step.startsWith("@")) {
                // This must be an element selector.
                // now split the step into element, position, attribute name, value, etc.
                if (!step.matches("[^\\[\\s]+((\\[\\d+\\])(\\[@[^\\s=]+=\"[^\"]*\"\\])??|(\\[@[^\\s=]+=\"[^\"]*\"\\]))??")) {
                    response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Syntax error in node selector step " + step);
                    return null;
                }

                String params[] = step.split("[\\[\\]]", 6);
                String elementName = params[0];
                
                Vector nodes = new Vector();
                NodeList children = node.getChildNodes();
                parent = node;
                
                // for each of the child found, compare it with the path element
                for (int k=0, posCount=0; k<children.getLength(); ++k) { 
                   Node child = children.item(k);
                   if (child.getNodeType() == Node.ELEMENT_NODE) {
                       Element el = (Element) child;
                       if (elementName.equals("*") || elementName.equals(el.getTagName())) { 
                           // if match found, check that child for the right position, attribute etc

                            int elementPos = -1;
                            
                            int posIndex  = 1, attrIndex = 1;
                            if (params.length > posIndex && !params[posIndex].startsWith("@")) {
                                // then its position
                                try {
                                    elementPos = Integer.parseInt(params[1]);
                                }
                                catch (NumberFormatException nfe) {// occurs if the part cannnot be parsed, if it is not Integer
                                    response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Position " + params[1] + " is not an integer");                           
                                    return null;
                                }
                                attrIndex += 2;
                                
                                if (posCount != elementPos) {
                                    posCount++;
                                    continue; //go to the next element.
                                }
                                else {
                                    posCount++;
                                }
                            }
                            
                            if (params.length > attrIndex && params[attrIndex].startsWith("@")) {
                                String hv = params[attrIndex];
                                String attrParts[] = hv.substring(1, hv.length()).split("=");
                                String attrName = attrParts[0];
                                String attrValue = attrParts[1];
                                
                                if (!el.hasAttribute(attrName)) {
                                    if (elementPos >= 0) {
                                        response.sendError(HttpServletResponse.SC_NOT_FOUND, "Attribute " + attrName + " not found in the element.");
                                        return null;
                                    }
                                    else {
                                        continue;
                                    }
                                }
                                String attr = el.getAttribute(attrName);
                                if (!attrValue.equals("\"" + attr + "\"")) {
                                    if (elementPos >= 0) {
                                        response.sendError(HttpServletResponse.SC_NOT_FOUND, "Attribute " + attrName + " does not match in the element");
                                        return null;
                                    }
                                    else {
                                        continue;
                                    }
                                }
                            }
                           
                            node = child;
                            nodes.add(step);
                       }
                       else {
                           errorStep = step;
                       }
                   }
                }
                
                if (nodes.size() > 1) {
                    response.sendError(HttpServletResponse.SC_MULTIPLE_CHOICES, "Many matches found for " + errorStep);
                    return null;
                }
                if (nodes.size() == 0) { // break the loop if some intermediate element was not found
                    if (j == steps.length - 1) {
                        //add this element as child of original_node
                        node = null;
                        break;
                    }
                    else {
                      response.sendError(HttpServletResponse.SC_NOT_FOUND, "Node selector not found for " + errorStep);
                      return null;
                    }
                }
            }
            else {
                // this is an attribute selector.
                //should also check that this is the last step
                if (j != steps.length - 1) {
                    response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Attribute selector must be the last step");
                    return null;
                }
                String attrName = step.substring(1, step.length());
                Element el = (Element) node;
                parent = node;
                if (el.hasAttribute(attrName)) {
                    node = el.getAttributeNode(attrName);
                }    
                else {
                    node = null;
                }
                break;
            }
        }
        
        Node[] nodes = new Node[2];
        nodes[0] = parent;
        nodes[1] = node;
        return nodes;
    }
    
    /** Handles the HTTP <code>GET</code> method.
     * @param request servlet request
     * @param response servlet response
     */
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
      try {
          doGetInternal(request, response);
      }
      catch (NoSuchAlgorithmException e) {
        response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,"Can not find MD5");
        return;
      }
    }
    
    protected void doGetInternal(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException, NoSuchAlgorithmException {
        // disable the browser cache 
        disableCache(response);
        
        // get the enitre path to the document or element
        String docNode = request.getPathInfo();
        // if the path not found
        if (docNode == null) {
            response.sendError(HttpServletResponse.SC_NOT_FOUND,"Document not specified");
            return;
        }
        // if the path found split it into document selector and node selector
        String parts[] = docNode.split("/~~/");
        String docPath = parts[0]; // the 1st part for document path
        String docStep[] = docPath.split("/"); // split the doc path on / to get application name..used for response type
        String filePath = ROOT_DIR + docPath; // append the root dir to filepath, hardcoded for time being
        
        String user = null;
        if (docStep.length >= 4 && docStep[2].equals("users")) {
            user = new String(docStep[3]);
            if (!authenticateDigest(request, response, user)) {
                return;
            }
        }
        else {
            /* This is for global. So no need to authenticate for GET */
            if (!authenticateDigest(request, response, "guest")) {
                return;
            }
        }
          

        File document = new File(filePath);// open the file
        if (document.isDirectory()) { // check if the path is global directory, if yes, send the index file
            filePath = filePath + "/" + INDEX_FILE;
            document = new File(filePath);
        }
        
        if (!document.exists()) { // if the file does not exist
            response.sendError(HttpServletResponse.SC_NOT_FOUND, "Document not found");
            return;
        }
        
        if (!document.canRead()) { // if file cannot be read
            response.sendError(HttpServletResponse.SC_NOT_FOUND, "Document not readable");
            return;
        }    
        if (parts.length > 2) { // if more than one ~~
            response.sendError(400, "Multiple node selectors not supported");
            return;
        }

        boolean if_match_present = request.getHeaders("If-Match").hasMoreElements();
        boolean if_none_match_present = request.getHeaders("If-None-Match").hasMoreElements();
        
        if (parts.length == 1) { // if only document selector found,read the document n send it
            String etag = calcEntityTag(document);
            setEntityTag(response, etag);
            setLastModified(response, document);
                        
            // this is as per specification..check
            if (matchEntityTag(request,response,document,etag)) {
              // no-entity tag was set, so return the document
              doGetDocument(response,docStep[1],document);  
              return;
            }
            else {
              if (if_none_match_present) {
                response.sendError(HttpServletResponse.SC_NOT_MODIFIED); 
              }
              else if (if_match_present) {
                response.sendError(HttpServletResponse.SC_PRECONDITION_FAILED); 
              }                  
              return;
            }
        }

        //parts is more than one, so one node selector is present.        
        String nodeSelector = parts[1];

        if (nodeSelector == null || nodeSelector.equals("")) { // ask
            response.sendError(HttpServletResponse.SC_NOT_FOUND,"No node selector specified");
            return;
        }    

        Document theDocument;
        try {
            String schemafile = getSchemaFileName(docStep[1]);
            theDocument = getParsedDocument(schemafile, new InputSource(new FileReader(filePath)));
        }
        catch (SAXException e) {
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Error parsing the document");
            return;
        }

        Node node = getNodeFromSelector(theDocument, nodeSelector, response);
        if (node == null) {
            return;
        }
        
        //construct the document with one node only.
        NodeList children = theDocument.getChildNodes(); // now remove all the node from document
        for (int j = children.getLength() - 1; j >= 0; --j) {
            theDocument.removeChild(children.item(j));
        }
        
        if (node != null) {
            String etag = calcEntityTag(document);
            if (!matchEntityTag(request,response,document,etag)) {
                setLastModified(response, document);
                setEntityTag(response, etag);
                response.sendError(HttpServletResponse.SC_NOT_MODIFIED,"The requested resource has not been modified "); 
                return;
            }
            
            // continue in case the if-none-match has succeded or if-match has failed or if there was no-entity tag
            if (node.getNodeType() == Node.ATTRIBUTE_NODE) {
                Attr attr = (Attr) node;
                // Output the XML
                setLastModified(response, document);
                setEntityTag(response, etag);
                response.setContentType("application/xcap-att+xml");
                PrintWriter out = response.getWriter(); // will be used only for error response
                out.println("\"" + attr.getValue() + "\""); 
                out.close();             
            }
            else if (node.getNodeType() == Node.ELEMENT_NODE) {
                theDocument.appendChild(node); // add the found node

                // Serialize the selected document to a string
                OutputFormat format  = new OutputFormat(theDocument);   
                StringWriter stringOut = new StringWriter();        
                XMLSerializer serial = new XMLSerializer( stringOut, format );
                serial.asDOMSerializer();                            
                serial.serialize( theDocument.getDocumentElement());

                // Output the XML
                setLastModified(response, document);
                setEntityTag(response, etag);
                response.setContentType("application/xcap-el+xml");
                PrintWriter out = response.getWriter(); // will be used only for error response
                String str = stringOut.toString();
                String s = str.replaceFirst("<[^>]*>\r?\n", ""); // eliminate the blank lines which are also treated as child nodes
                out.println(s); 
                out.close();             
            }
            else {
                response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Not an element or attribute");
                return;
            }
        }
        else {
            response.sendError(HttpServletResponse.SC_NOT_FOUND);
            return;
        }
    }

     /** Handles the HTTP <code>PUT</code> method.
     * @param request servlet request
     * @param response servlet response
     */
    protected void doPut(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
      try {
          doPutInternal(request, response);
      }
      catch (NoSuchAlgorithmException e) {
        response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,"Can not find MD5");
        return;
      }
    }
    
    protected void doPutInternal(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException, NoSuchAlgorithmException {  
        // get the enitre path to the document or element
        String docNode = request.getPathInfo();
        // if the path not found
        if (docNode == null) {
            response.sendError(HttpServletResponse.SC_NOT_FOUND,"Document not specified");
            return;
        }
        // if the path found split it into document selector and node selector
        String parts[] = docNode.split("/~~/");
        String docPath = parts[0]; // the 1st part for document path
        String docStep[] = docPath.split("/"); // split the doc path on / to get application name..used for response type
        String filePath = ROOT_DIR + docPath; // append the root dir to filepath, hardcoded for time being
            
        File document = new File(filePath);// open the file
       
        String user = null;
        if (docStep.length >= 4 && docStep[2].equals("users")) {
            user = new String(docStep[3]);
            if (!authenticateDigest(request, response, user)) {
                return;
            }
        }
        else {
            if (!authenticateDigest(request, response, "admin")) {
                return;
            }
        }
        
        // the directory has to exist as per specification, so we check that first
        File dir = document.getParentFile();
        if (dir == null) {
           response.sendError(HttpServletResponse.SC_CONFLICT, "Cannot locate directory " +dir+" in "+docPath);
           return;
        }
            
        if (!dir.exists()) {
           response.sendError(HttpServletResponse.SC_CONFLICT, "Cannot locate directory " +dir+" in "+docPath );
           return;
        }
        
        if (document.isDirectory()) { // check if the path is directory, if yes, we are trying to put index file
            filePath = filePath + "/" + INDEX_FILE;
            document = new File(filePath);
        }
        
        if (parts.length > 1 && !document.exists()) { 
            //Node selector is specified but document doesn't exist
            response.sendError(HttpServletResponse.SC_CONFLICT, "document " + docPath +" not found"); 
            return;
        }
        
        if (parts.length > 2) { // if more than one ~~
            response.sendError(400, "Multiple node selectors not supported");
            return;
        }
        
        // before creating the document we shud check if its parent exist
        if (!document.exists() && parts.length == 1) { 
             // if the document does not exist
            // check if there were any if-match headers, if none-match headers supplied
            // if "if-match" was supplied, it will never succeed, so we have to say document does not exist
            Enumeration if_match = request.getHeaders("If-Match");
            
            if (if_match.hasMoreElements()) {
                response.sendError(HttpServletResponse.SC_NOT_FOUND, "Precondition present for non-existent document");
                return;
            }
            
            // if it a if-none-match header, the request should succeed n new document has to be created, so no need to check
           
            
            if (document.createNewFile() == false) {
                response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Can not create file");
                return;
            }
            
            if (!document.canWrite()) { // if file cannot be read
                response.sendError(HttpServletResponse.SC_NOT_FOUND, "Document not writable");
                return;
            }            
            // if everything is fine...put the document            
            doPutDocument(request, response, docStep[1], document);
            //etag = calcDigest(response,document);
            setLastModified(response, document);
            setEntityTag(response, document); 
            return;
        }

        if (document.exists() && parts.length == 1) { // if only document selector found
            //System.out.println("putting the document");
            String etag = calcEntityTag(document);
            if (matchEntityTag(request,response,document,etag)) { // its a replacement case, whole doc is being replaced     
              doPutDocument(request,response,docStep[1],document);
              setLastModified(response,  document);
              setEntityTag(response, document);   
            }
            else {
              setLastModified(response, document);  
              setEntityTag(response,etag);
              response.sendError(HttpServletResponse.SC_PRECONDITION_FAILED);
            }
            return;            
        }

        //parts is more than one, so one node selector is present n document exists...        
        String nodeSelector = parts[1];

        if (nodeSelector == null || nodeSelector.equals("")) { // ask
            response.sendError(HttpServletResponse.SC_CONFLICT,"No node selector specified");
            return;
        }    

        Document theDocument;
        try {
            String schemafile = getSchemaFileName(docStep[1]);
            theDocument = getParsedDocument(schemafile, new InputSource(new FileReader(filePath)));
        }
        catch (SAXException e) {
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Error parsing the document");
            return;
        }

        Node[] nodes = getNodeFromSelector2(theDocument, nodeSelector, response);
        if (nodes == null) {
            return;
        }
        
        Node parent = nodes[0];
        Node node = nodes[1];
        
        if (parent == null) {
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Parent node is null");
            return;
        }
        if (parent.getNodeType() != Node.ELEMENT_NODE) {
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Parent node is not element");
            return;
        }
            
        //two cases: node is null (insert) or not null (replace)
        String content_type = request.getContentType();
        String steps[] = nodeSelector.split("/");
        if (content_type != null && 
        content_type.equalsIgnoreCase("application/xcap-el+xml") &&
        !steps[steps.length - 1].startsWith("@")) {
            String body = new String("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\r\n"); 
            BufferedReader in = request.getReader();
            String str;
            while ((str = in.readLine()) != null) {
                body += str;
                body += "\r\n";
            }
            in.close();

            // Create and configure a DOM parser
            DOMParser body_parser = new DOMParser();

            // configure the parser
            configParser(body_parser);

            // Parse the document
            try {
               StringReader bodyStream = new StringReader(body);
               System.out.println("Parsing \n" + body);
               body_parser.parse(new InputSource(bodyStream));
            }
            catch (SAXException e) {
                response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Error parsing the XML element");
                return;
            }
            catch (Exception e) {
                e.printStackTrace();
                response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Error in the original document");
                return;
            }

            Document theElement = body_parser.getDocument(); // the document

            //now check if the supplied element is a single element or not.
            NodeList children = theElement.getChildNodes();
            int count = 0;
            Node theNode = null;
            for (int j = children.getLength() - 1; j >= 0; --j) {
                if (children.item(j).getNodeType() == Node.ELEMENT_NODE) {
                    count ++;
                    theNode = children.item(j);
                }
            }

            if (count != 1 || theNode == null) {
                response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Found " + count + " elements");
                return;
            }

            /* check for attribute consistency */
            String params[] = steps[steps.length-1].split("[\\[\\]]", 6);
            String[] attrParts = new String[2];
            
            if (params.length >= 2 && params[1].indexOf('=') >= 0) {
               attrParts = params[1].substring(1, params[1].length()).split("=");
            }
            else if (params.length >= 4 && params[3].indexOf('=') >= 0) {
               attrParts = params[3].substring(1, params[3].length()).split("=");
            }
            else {
                attrParts[0] = null;
                attrParts[1] = null;
            }
            
            if (attrParts[0] != null) {
                /* This must match the attribute name in the element */ 
                Element topElement = theElement.getDocumentElement();
                String attr = topElement.getAttribute(attrParts[0]);
                if (!attrParts[1].equals("\"" + attr + "\"")) {
                    response.sendError(HttpServletResponse.SC_CONFLICT, "Attribute in URI and body do not match: " + attrParts[1] + "!=\"" + attr + "\"");
                    return;
                }
            }
            
            if (node == null) { // its an addition
                Enumeration if_match = request.getHeaders("If-Match");
                if (if_match.hasMoreElements()) { // if there is a if-match header..
                    response.sendError(HttpServletResponse.SC_PRECONDITION_FAILED);
                    return;
                }
                // if-none-match will always succeed so no check for that
                Node n = theDocument.importNode(theNode, true);
                parent.appendChild(n);
            }
            else {
                // it is a replacement
                // here we check the entity tags b4 replacing
              String etag = calcEntityTag(document);  
              if (matchEntityTag(request, response, document, etag)) {        
                 Node n = theDocument.importNode(theNode, true);
                 System.out.println("Replacing in " + parent + " node=" + node + " to new node=" + n);
                 parent.replaceChild(n, node); 
              }
              else {
                 setLastModified(response, document); 
                 setEntityTag(response, etag);
                 response.sendError(HttpServletResponse.SC_PRECONDITION_FAILED);
                 return;
              }              
            }
        }
        else if (content_type != null &&content_type.equalsIgnoreCase("application/xcap-att+xml") &&
        steps[steps.length - 1].startsWith("@")) {
            //this is attribute
            Element parentElement = (Element) parent;
            if (node != null && node.getNodeType() == Node.ATTRIBUTE_NODE) {
               String etag = calcEntityTag(document);
               if (!matchEntityTag(request, response, document, etag)) {
                 setLastModified(response, document);  
                 setEntityTag(response, etag);
                 response.sendError(HttpServletResponse.SC_PRECONDITION_FAILED);
                 return;
               }
               else {
                parentElement.removeAttributeNode((Attr) node);  
                // rest of the code is common if the node is null or not-null
               }
            }
            else if (node == null) {
                 Enumeration if_match = request.getHeaders("If-Match");
                 if (if_match.hasMoreElements()) { // if there is a if-match header..
                        response.sendError(HttpServletResponse.SC_PRECONDITION_FAILED);
                        return;
                    // if-none-match will always succeed so no check for that
                } 
            }    
            //insert a new attribute
            String attr = steps[steps.length - 1].substring(1);
            BufferedReader in = request.getReader();
            String str;
            String body = new String("");
            while ((str = in.readLine()) != null) {
                body += str;
                body += "\r\n";
            }
            in.close();
            body.trim(); // remove whitespaces.
            String value;
            if (body.startsWith("\"") && body.endsWith("\"")) {
               value = body.substring(1, body.length() - 1);
            }
            else {
                value = body;
            }
            parentElement.setAttribute(attr, value);
        }
        else {
            String type;
            if (steps[steps.length - 1].startsWith("@")) {
                type = "application/xcap-att+xml";
            }
            else {
                type = "application/xcap-el+xml";
            }
            response.setHeader("Accept", type);
            response.sendError(HttpServletResponse.SC_UNSUPPORTED_MEDIA_TYPE, "Content type is incorrect");
            return;
        }
        
        // normalize the document
        // theDocument.normalize();
        
        // Serialize the selected document to a string
        OutputFormat format  = new OutputFormat(theDocument);
        StringWriter stringOut = new StringWriter();        
        XMLSerializer serial = new XMLSerializer( stringOut, format );
        serial.asDOMSerializer();                            
        serial.serialize( theDocument.getDocumentElement());
        String newDocument = stringOut.toString();
        
        //parse the new document for schema validation.
        try {
           theDocument = getParsedDocument(getSchemaFileName(docStep[1]), new InputSource(new StringReader(newDocument)));
        }
        catch (SAXException e) {
            response.sendError(HttpServletResponse.SC_CONFLICT, "Parsing error: " + e.getMessage());
            return;            
        }
        
        // now update the whole document
        try {
            PrintWriter out = new PrintWriter(new FileWriter(document));
            out.write(newDocument);
            // everything is done, the document has been modified in whtever way poss..so set the etag
            setLastModified(response, document);
            setEntityTag(response, document);        
            out.close();
        } catch (IOException e) {
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Error writing file: " + e.getMessage());
        }                
    }
   
     /** Handles the HTTP <code>DELETE</code> method.
     * @param request servlet request
     * @param response servlet response
     */
    protected void doDelete(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
      try {
          doDeleteInternal(request, response);
      }
      catch (NoSuchAlgorithmException e) {
        response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,"Can not find MD5");
        return;
      }
    }
    
    protected void doDeleteInternal(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException, NoSuchAlgorithmException {

        String docNode = request.getPathInfo();
        // if the path not found
        if (docNode == null) {
            response.sendError(HttpServletResponse.SC_NOT_FOUND,"Document not specified");
            return;
        }
        // if the path found split it into document selector and node selector
        String parts[] = docNode.split("/~~/");
        String docPath = parts[0]; // the 1st part for document path
        String docStep[] = docPath.split("/"); // split the doc path on / to get application name..used for response type
        String filePath = ROOT_DIR + docPath; // append the root dir to filepath, hardcoded for time being
            
        // Get Authorization header
        String user = null;
        if (docStep.length >= 4 && docStep[2].equals("users")) {
            user = new String(docStep[3]);
            if (!authenticateDigest(request, response, user)) {
                return;
            }
        }
        else {
            if (!authenticateDigest(request, response, "admin")) {
                return;
            }
        }
        
        File document = new File(filePath);// open the file
        if (document.isDirectory()) { // check if the path is directory, if yes, send the index file
            filePath = filePath + "/" + INDEX_FILE;
            document = new File(filePath);
        }

        if (parts.length > 1 && !document.exists()) { 
            //Node selector is specified but document doesn't exist
            response.sendError(HttpServletResponse.SC_NOT_FOUND, "Document not found"); 
            return;
        }
        
        /////////-----------------Check, shud i do entity tags for this????------------------------
        
        if (!document.exists()) { // if the file does not exist
            // does not exist then create
            response.setStatus(HttpServletResponse.SC_OK);
            return;
        }
        
        if (!document.canWrite()) { // if file cannot be written
            response.sendError(HttpServletResponse.SC_NOT_FOUND, "Document not writable");
            return;
        }
        if (parts.length > 2) { // if more than one ~~
            response.sendError(400, "Multiple node selectors not supported");
            return;
        }

        if (parts.length == 1) { // if only document selector found,read the document n send it
            String etag = calcEntityTag(document);
            if (matchEntityTag(request,response,document,etag)) {
                document.delete();
                response.setStatus(HttpServletResponse.SC_OK);
            }    
            else {
              //setLastModified(response, document);  
              //setEntityTag(response, etag);
              response.sendError(HttpServletResponse.SC_PRECONDITION_FAILED);
            }
            return;
        }

        //parts is more than one, so one node selector is present.        
        String nodeSelector = parts[1];

        if (nodeSelector == null || nodeSelector.equals("")) { // ask
            response.sendError(HttpServletResponse.SC_NOT_FOUND,"No node selector specified");
            return;
        }    

        Document theDocument;
        try {
            String schemafile = getSchemaFileName(docStep[1]);
            theDocument = getParsedDocument(schemafile, new InputSource(new FileReader(filePath)));
        }
        catch (SAXException e) {
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Error parsing the document");
            return;
        }

        Node[] nodes = getNodeFromSelector3(theDocument, nodeSelector, response);
        if (nodes == null) {
            return;
        }
        Node parent = nodes[0];
        Node node = nodes[1];
        
        if (parent == null) {
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Parent node is null");
            return;
        }
        if (parent.getNodeType() != Node.ELEMENT_NODE) {
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Parent node is not element");
            return;
        }
            
        //two cases: node is null (no operation) or not null (delete)
        String steps[] = nodeSelector.split("/");
        if (!steps[steps.length - 1].startsWith("@")) {
            if (node != null) {
                String etag = calcEntityTag(document);
                if (!matchEntityTag(request,response,document,etag)) {
                    //setLastModified(response, document);
                    //setEntityTag(response, etag);
                    response.sendError(HttpServletResponse.SC_PRECONDITION_FAILED);
                    return;
                }
                else {
                    parent.removeChild(node);
                    //setEntityTag(response, document);                    
                }
            }
        }
        else if (steps[steps.length - 1].startsWith("@")) {
            Element parentElement = (Element) parent;
            if (node != null && node.getNodeType() == Node.ATTRIBUTE_NODE) {
                String etag = calcEntityTag(document);
                if (!matchEntityTag(request,response,document,etag)) {
                    //setLastModified(response, document);
                    //setEntityTag(response, etag);
                    response.sendError(HttpServletResponse.SC_PRECONDITION_FAILED);
                    return;
                }
                else {
                    parentElement.removeAttributeNode((Attr) node);
                    //setEntityTag(response, document);
                }    
            }
        }
        
        // normalize the document
        // theDocument.normalize();
        
        // Serialize the selected document to a string
        OutputFormat format  = new OutputFormat(theDocument);
        StringWriter stringOut = new StringWriter();        
        XMLSerializer serial = new XMLSerializer( stringOut, format );
        serial.asDOMSerializer();                            
        serial.serialize( theDocument.getDocumentElement());

        String newDocument = stringOut.toString();
        
        //parse the new document for schema validation.
        try {
           theDocument = getParsedDocument(getSchemaFileName(docStep[1]), new InputSource(new StringReader(newDocument)));
        }
        catch (SAXException e) {
            response.sendError(HttpServletResponse.SC_CONFLICT, "Parsing error: " + e.getMessage());
            return;            
        }
        
        // now update the whole document
        try {
            PrintWriter out = new PrintWriter(new FileWriter(document));
            out.write(newDocument);
            // the document has chenged in someway..return the changed etag
            //setLastModified(response, document);
            //setEntityTag(response, document);
            out.close();
        } catch (IOException e) {
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Error writing file: " + e.getMessage());
            return;
        }                
        
        response.setStatus(HttpServletResponse.SC_OK);        
    }
   
     /** Handles the HTTP <code>POST</code> method.
     * @param request servlet request
     * @param response servlet response
     */
    protected void doPost(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {   
        
        //doPut(request, response);
        response.sendError(HttpServletResponse.SC_METHOD_NOT_ALLOWED, "POST method not allowed for XCAP");
        /* output your page here*/
    }
   
    /** Returns a short description of the servlet.
     */
    public String getServletInfo() {
        return "XCAP Server (c) 2004 Columbia University";
    }
    
    /** Disable cache for the response.
     */
    public void disableCache(HttpServletResponse response) {
        response.setHeader("Pragma", "No-cache");
        response.setHeader("Cache-Control", "no-cache");
        response.setDateHeader("Expires", 0);        
    }
    
    /** Authentication. */
    protected boolean allowUser(String login, String auth) 
    throws IOException {
        if (auth == null) return false;  // no auth

        if (auth.toUpperCase().startsWith("BASIC ")) { 
            return false; //we do not allow BASIC authentication. always fail.
            /*
            // Get encoded user and password, comes after "BASIC "
            String userpassEncoded = auth.substring(6);

            // Decode it, using any base 64 decoder
            sun.misc.BASE64Decoder dec = new sun.misc.BASE64Decoder();
            String userpassDecoded = new String(dec.decodeBuffer(userpassEncoded));

            // Check our user list to see if that user and password are "allowed"
            if (userpassDecoded.equals(login + ":" + login))
              return true;
            else
              return false;
            */        
        }
        else if (auth.toUpperCase().startsWith("DIGEST ")) {
 
            //System.out.println("Auth " + auth);
            if (login.equals("guest")) {
                //System.out.println("Guest is allowed always.");
                return true; //TODO: Is this alright?
            }
            
            String[] parts = auth.split("[\\s,\\r\\n]+");
            String username = null;
            for (int i=0; i<parts.length; ++i) {
                //System.out.println("parts[" + i + "] = " + parts[i]);
                if (parts[i].matches("^username=.*")) {
                    username = new String(parts[i].substring(10, parts[i].length()-1));
                    //System.out.println("username = " + username);
                    break;
                }
            }
            if (username != null && (username.equals("admin") || username.equals(login))) {
                //System.out.println("Username is " + username + ", login is " + login);
                return true;
            }
            //System.out.println("Auth failed for username " + username + " and login " + login);
            return false;
        }
        return false;
    }
    
    /**
     * try to authenticate using Basic. If the authentication fails it sends the error response.
     * Return true on success and false on failure.
     */
    protected boolean authenticateBasic(HttpServletRequest request, HttpServletResponse response, String user)
    throws IOException {
        String auth = request.getHeader("Authorization");
        if (auth == null || !allowUser(user, auth)) {
            // Not allowed, so report he's unauthorized
            response.setHeader("WWW-Authenticate", "BASIC realm=\"XCAP server access\"");
            response.sendError(HttpServletResponse.SC_UNAUTHORIZED);
            return false;
        }
        return true;
    }
    
    protected boolean authenticateDigest(HttpServletRequest request, 
                                HttpServletResponse response, String user)
    throws IOException {
        String auth = request.getHeader("Authorization");
        if (auth == null || !allowUser(user, auth)) {
            // Not allowed, so report he's unauthorized
            response.setHeader("WWW-Authenticate", "DIGEST realm=\"XCAPServer\"");
            response.sendError(HttpServletResponse.SC_UNAUTHORIZED,
                               "Access denied. Requires user=" + user);
            return false;
        }
        return true;
    }
    
    // Returns the contents of the file in a byte array.
    public static byte[] getBytesFromFile(File file) throws IOException {
        InputStream is = new FileInputStream(file);
    
        // Get the size of the file
        long length = file.length();
    
        // You cannot create an array using a long type.
        // It needs to be an int type.
        // Before converting to an int type, check
        // to ensure that file is not larger than Integer.MAX_VALUE.
        if (length > Integer.MAX_VALUE) {
            // File is too large
            //???? how to handle this situation
        }
    
        // Create the byte array to hold the data
        byte[] bytes = new byte[(int)length];
    
        // Read in the bytes
        int offset = 0;
        int numRead = 0;
        while (offset < bytes.length
               && (numRead=is.read(bytes, offset, bytes.length-offset)) >= 0) {
            offset += numRead;
        }
    
        // Ensure all the bytes have been read in
        if (offset < bytes.length) {
            throw new IOException("Could not completely read file "+file.getName());
        }
    
        // Close the input stream and return bytes
        is.close();
        return bytes;
    }

    protected String calcDigest(File file) throws IOException, NoSuchAlgorithmException {
        // Create a Message Digest from a Factory method
        MessageDigest md = MessageDigest.getInstance("MD5");

        // Create the message, convert the file contents to an array of bytes[]
        byte[] msg = getBytesFromFile(file);

        // Update the message digest with some more bytes
        // This can be performed multiple times before creating the hash
        md.update(msg); 

        // Create the digest from the message
        byte[] aMessageDigest = md.digest();
        sun.misc.BASE64Encoder enc = new sun.misc.BASE64Encoder();

        String encodedDigest = new String(enc.encode(aMessageDigest));
        // Printout for debugging
        // System.out.println("Original: " + new String(msg));
        System.out.println("Message Digest: " + new String(encodedDigest));        
        return(new String(encodedDigest));
    }    
    
    protected String calcEntityTag(File file) throws IOException, NoSuchAlgorithmException {
        return calcDigest(file);
    }
    
    /** Set entity tag */
    protected void setEntityTag(HttpServletResponse response, File file) 
    throws IOException, NoSuchAlgorithmException {       
      String etag = calcDigest(file);    
      response.setHeader("ETag", "\"" + etag + "\"");
    }
    
    /** Set entity tag */
    protected void setEntityTag(HttpServletResponse response, String etag) throws IOException
    {       
        response.setHeader("ETag", "\"" + etag + "\"");
    }
    
    
    protected void setLastModified (HttpServletResponse response, File file) {
        long last_modified = file.lastModified();
        //Last-Modified: Wed, 15 Nov 1995 04:58:08 GMT
        //response.setHeader("Last-Modified", "\"" + String.valueOf(last_modified) + "\"");
    }
    
    /** Match the If-Match and If-None-Match header for this document. */
    protected boolean matchEntityTag(HttpServletRequest request, HttpServletResponse response, File file, String etag) throws IOException
    {
        System.out.println("In matchEntityTag: "+etag);
        Enumeration if_match = request.getHeaders("If-Match");
        if (if_match.hasMoreElements()) {
            do {
                String v = (String) if_match.nextElement();
                String n[] = v.split(",");
                for (int i=0; i<n.length; ++i) {
                    String value = n[i].trim();
                    if (value.equals("\"*\"") || value.matches("(W/)?\"" + etag + "\"")) {
                      return true;
                    }
                }
            } while (if_match.hasMoreElements());
            return false;
        }
        
        Enumeration if_none_match = request.getHeaders("If-None-Match");
        if (if_none_match.hasMoreElements()) {
            do {
                String v = (String) if_none_match.nextElement();
                String n[] = v.split(",");
                for (int i=0; i<n.length; ++i) {
                    String value = n[i].trim();
                    if (value.equals("\"*\"") || value.matches("(W/)?\"" + etag + "\"")) {
                      return false;
                    }
                }
            } while (if_none_match.hasMoreElements());
            return true;
        }
        
        return true;
    }
}
